"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const core_1 = require("@stylable/core");
const vscode_languageserver_textdocument_1 = require("vscode-languageserver-textdocument");
const vscode_languageserver_types_1 = require("vscode-languageserver-types");
const vscode_uri_1 = require("vscode-uri");
const completion_providers_1 = require("./completion-providers");
const css_service_1 = require("./css-service");
const dedupe_refs_1 = require("./dedupe-refs");
const diagnosis_1 = require("./diagnosis");
const color_provider_1 = require("./feature/color-provider");
const provider_1 = require("./provider");
const provider_2 = require("./provider");
const typescript_support_1 = require("./typescript-support");
class StylableLanguageService {
    constructor({ fs, stylable }) {
        this.fs = fs;
        this.stylable = stylable;
        this.tsLanguageService = typescript_support_1.typescriptSupport(this.fs);
        this.provider = new provider_1.Provider(this.stylable, this.tsLanguageService);
        this.cssService = new css_service_1.CssService(this.fs);
    }
    getStylable() {
        return this.stylable;
    }
    getFs() {
        return this.fs;
    }
    onCompletion(filePath, offset) {
        const stylableFile = this.readStylableFile(filePath);
        if (stylableFile && stylableFile.stat.isFile()) {
            const document = vscode_languageserver_textdocument_1.TextDocument.create(vscode_uri_1.URI.file(filePath).toString(), 'stylable', stylableFile.stat.mtime.getTime(), stylableFile.content);
            const position = document.positionAt(offset);
            return this.getCompletions(document, filePath, position);
        }
        else {
            return [];
        }
    }
    onDefinition(filePath, offset) {
        const stylableFile = this.readStylableFile(filePath);
        if (stylableFile && stylableFile.stat.isFile()) {
            const doc = vscode_languageserver_textdocument_1.TextDocument.create(vscode_uri_1.URI.file(filePath).toString(), 'stylable', stylableFile.stat.mtime.getTime(), stylableFile.content);
            const res = this.provider.getDefinitionLocation(stylableFile.content, doc.positionAt(offset), filePath, this.fs);
            return res.map((loc) => vscode_languageserver_types_1.Location.create(vscode_uri_1.URI.file(loc.uri).toString(), loc.range));
        }
        return [];
    }
    onHover(filePath, offset) {
        const stylableFile = this.readStylableFile(filePath);
        if (stylableFile && stylableFile.stat.isFile()) {
            const doc = vscode_languageserver_textdocument_1.TextDocument.create(vscode_uri_1.URI.file(filePath).toString(), 'stylable', stylableFile.stat.mtime.getTime(), stylableFile.content);
            return this.cssService.doHover(doc, doc.positionAt(offset));
        }
        return null;
    }
    onReferences(filePath, offset) {
        const stylableFile = this.readStylableFile(filePath);
        if (stylableFile && stylableFile.stat.isFile()) {
            const doc = vscode_languageserver_textdocument_1.TextDocument.create(vscode_uri_1.URI.file(filePath).toString(), 'stylable', stylableFile.stat.mtime.getTime(), stylableFile.content);
            const position = doc.positionAt(offset);
            const refs = this.getRefs(filePath, position);
            if (refs.length) {
                return dedupe_refs_1.dedupeRefs(refs);
            }
            else {
                return dedupe_refs_1.dedupeRefs(this.cssService.findReferences(doc, position));
            }
        }
        return [];
    }
    onDocumentColor(filePath) {
        const stylableFile = this.readStylableFile(filePath);
        if (stylableFile && stylableFile.stat.isFile()) {
            const doc = vscode_languageserver_textdocument_1.TextDocument.create(vscode_uri_1.URI.file(filePath).toString(), 'stylable', stylableFile.stat.mtime.getTime(), stylableFile.content);
            return color_provider_1.resolveDocumentColors(this.stylable, this.cssService, doc, this.fs);
        }
        return [];
    }
    onColorPresentation(filePath, offset, color) {
        const stylableFile = this.readStylableFile(filePath);
        if (stylableFile && stylableFile.stat.isFile()) {
            const doc = vscode_languageserver_textdocument_1.TextDocument.create(vscode_uri_1.URI.file(filePath).toString(), 'stylable', stylableFile.stat.mtime.getTime(), stylableFile.content);
            const range = vscode_languageserver_types_1.Range.create(doc.positionAt(offset.start), doc.positionAt(offset.end));
            return color_provider_1.getColorPresentation(this.cssService, doc, { color, range, textDocument: doc });
        }
        return [];
    }
    onRenameRequest(filePath, offset, newName) {
        const stylableFile = this.readStylableFile(filePath);
        const edit = { changes: {} };
        if (stylableFile && stylableFile.stat.isFile()) {
            const doc = vscode_languageserver_textdocument_1.TextDocument.create(vscode_uri_1.URI.file(filePath).toString(), 'stylable', stylableFile.stat.mtime.getTime(), stylableFile.content);
            provider_2.getRenameRefs(filePath, doc.positionAt(offset), this.fs, this.stylable).forEach((ref) => {
                if (edit.changes[ref.uri]) {
                    edit.changes[ref.uri].push({ range: ref.range, newText: newName });
                }
                else {
                    edit.changes[ref.uri] = [{ range: ref.range, newText: newName }];
                }
            });
        }
        return edit;
    }
    onSignatureHelp(filePath, offset) {
        const stylableFile = this.readStylableFile(filePath);
        if (stylableFile && stylableFile.stat.isFile()) {
            const doc = vscode_languageserver_textdocument_1.TextDocument.create(vscode_uri_1.URI.file(filePath).toString(), 'stylable', stylableFile.stat.mtime.getTime(), stylableFile.content);
            const sig = this.provider.getSignatureHelp(stylableFile.content, doc.positionAt(offset), filePath, this.fs, vscode_languageserver_types_1.ParameterInformation);
            return sig;
        }
        return null;
    }
    onDocumentFormatting() {
        // no op
        return null;
    }
    provideCompletionItemsFromSrc(src, pos, fileName) {
        return this.provider.provideCompletionItemsFromSrc(src, pos, fileName, this.fs);
    }
    getCompletions(document, filePath, position) {
        const content = document.getText();
        const res = this.provider.provideCompletionItemsFromSrc(content, {
            line: position.line,
            character: position.character,
        }, filePath, this.fs);
        const ast = core_1.safeParse(content, { from: filePath });
        const cleanDocument = this.cssService.createSanitizedDocument(ast, filePath, document.version);
        return res
            .map((com) => {
            const lspCompletion = vscode_languageserver_types_1.CompletionItem.create(com.label);
            const ted = vscode_languageserver_types_1.TextEdit.replace(com.range
                ? com.range
                : new completion_providers_1.ProviderRange(new completion_providers_1.ProviderPosition(position.line, Math.max(position.character - 1, 0)), position), typeof com.insertText === 'string' ? com.insertText : com.insertText.source);
            lspCompletion.insertTextFormat = 2;
            lspCompletion.detail = com.detail;
            lspCompletion.textEdit = ted;
            lspCompletion.sortText = com.sortText;
            lspCompletion.filterText =
                typeof com.insertText === 'string' ? com.insertText : com.insertText.source;
            if (com.additionalCompletions) {
                lspCompletion.command = vscode_languageserver_types_1.Command.create('additional', 'editor.action.triggerSuggest');
            }
            else if (com.triggerSignature) {
                lspCompletion.command = vscode_languageserver_types_1.Command.create('additional', 'editor.action.triggerParameterHints');
            }
            return lspCompletion;
        })
            .concat(this.cssService.getCompletions(cleanDocument, position));
    }
    getDefinitionLocation(src, position, filePath) {
        const defs = this.provider.getDefinitionLocation(src, position, vscode_uri_1.URI.file(filePath).fsPath, this.fs);
        return defs.map((loc) => vscode_languageserver_types_1.Location.create(vscode_uri_1.URI.file(loc.uri).fsPath, loc.range));
    }
    getSignatureHelp(src, pos, filePath, paramInfo) {
        return this.provider.getSignatureHelp(src, pos, filePath, this.fs, paramInfo);
    }
    getRefs(filePath, position) {
        return provider_2.getRefs(filePath, position, this.fs, this.stylable);
    }
    resolveDocumentColors(document) {
        return color_provider_1.resolveDocumentColors(this.stylable, this.cssService, document, this.fs);
    }
    getColorPresentation(document, params) {
        return color_provider_1.getColorPresentation(this.cssService, document, params);
    }
    diagnose(filePath) {
        const stylableFile = this.readStylableFile(filePath);
        if (stylableFile && stylableFile.stat.isFile()) {
            return diagnosis_1.createDiagnosis(stylableFile.content, stylableFile.stat.mtime.getTime(), filePath, this.stylable, this.cssService);
        }
        return [];
    }
    readStylableFile(filePath) {
        if (!filePath.endsWith('.st.css') && !filePath.startsWith('untitled:')) {
            return null;
        }
        let stat;
        try {
            stat = this.fs.statSync(filePath);
        }
        catch {
            // TODO: add warning?
        }
        if (stat && stat.isFile()) {
            const content = this.fs.readFileSync(filePath, 'utf8');
            return {
                content,
                stat,
            };
        }
        return null;
    }
}
exports.StylableLanguageService = StylableLanguageService;
//# sourceMappingURL=service.js.map